import unittest

from curlylint.rules.rule_test_case import RulesTestMeta

from .django_forms_rendering import django_forms_rendering


class TestRule(unittest.TestCase, metaclass=RulesTestMeta):
    fixtures = __file__.replace(".py", ".json")
    rule = django_forms_rendering



import sys
from pathlib import Path
from typing import Set

import parsy

from .check import check_file, check_files
from .file import File
from .issue import Issue, IssueLocation
from .parse import make_parser


def get_parsy_error_location(error, file_path):
    line, column = parsy.line_info_at(error.stream, error.index)
    return IssueLocation(line=line, column=column, file_path=file_path)


def parse_file(path_and_config):
    """
    Returns a tuple ([Issue], File | None).
    """
    path, config = path_and_config

    with path.open("r") as f:
        source = f.read()

    return parse_source(path, config, source)


def parse_source(path: Path, config, source: str):
    parser = make_parser(config)

    try:
        file = File(
            path=path,
            source=source,
            lines=source.split("\n"),
            tree=parser["content"].parse(source),
        )
        return [], file
    except parsy.ParseError as error:
        location = get_parsy_error_location(error, path)
        issue = Issue(location, "Parse error: " + str(error), "parse_error")
        return [issue], None


def lint(paths: Set[Path], config):
    issues = []
    files = []

    from multiprocessing import Pool

    pool = Pool()

    parse_file_args = ((p, config) for p in paths)
    results = pool.map(parse_file, parse_file_args)
    for result in results:
        parse_issues, file = result
        issues += parse_issues
        if file is not None:
            files.append(file)

    if config.get("parse_only", False):
        return issues

    rules = config.get("rules")

    if rules:
        issues += check_files(files, rules)

    return issues


def lint_one(path: Path, config):
    if not path.is_file() and str(path) == "-":
        source = sys.stdin.read()
        parse_issues, file = parse_source(
            config.get("stdin_filepath", path), config, source
        )
    else:
        parse_issues, file = parse_file((path, config))

    issues = parse_issues

    if config.get("parse_only", False):
        return issues

    rules = config.get("rules")

    if rules and file is not None:
        issues += check_file(file, rules)

    return issues



from itertools import groupby
from typing import List

import click

from curlylint.issue import Issue


def format_stylish(issues: List[Issue]):
    sorted_issues = sorted(
        issues,
        key=lambda i: (
            i.location.file_path,
            i.location.line,
            i.location.column,
        ),
    )

    for file_path, group in groupby(
        sorted_issues, lambda i: i.location.file_path
    ):
        print(click.style(str(file_path), underline=True))
        for issue in group:
            loc = click.style(
                f"{issue.location.line}:{issue.location.column}", dim=True
            )
            print(f"{loc}\t{issue.message}\t{issue.code}")
        print("")



import attr


@attr.s(frozen=True)
class Location:
    """A location in a source file."""

    line = attr.ib()  # 0-based
    column = attr.ib()
    index = attr.ib()

    def __str__(self):
        return "{}:{}".format(self.line + 1, self.column)


@attr.s(frozen=True)
class Node:
    """Base abstract type for AST nodes."""

    begin = attr.ib()  # Location of the first character of the node
    end = attr.ib()  # Location of the last character


@attr.s(frozen=True)
class Slash(Node):
    """The `/` of (self-)closing HTML tags"""

    def __str__(self):
        return "/"


@attr.s(frozen=True)
class OpeningTag(Node):
    name = attr.ib()
    attributes = attr.ib()  # Interpolated<Attribute>
    slash = attr.ib(default=None)  # Slash | None (`Slash` if self-closing tag)

    def __str__(self):
        name = str(self.name)
        inner = " ".join([name] + [str(a) for a in self.attributes])
        if self.slash:
            inner += " /"
        return "<" + inner + ">"


@attr.s(frozen=True)
class ClosingTag(Node):
    name = attr.ib()

    def __str__(self):
        return "</{}>".format(self.name)


@attr.s(frozen=True)
class Element(Node):
    opening_tag = attr.ib()  # OpeningTag
    closing_tag = attr.ib()  # ClosingTag | None
    content = attr.ib()  # Interpolated | None

    def __attrs_post_init__(self):
        assert (self.closing_tag is None) == (self.content is None)

        if self.closing_tag is not None:
            if isinstance(self.opening_tag.name, str):
                assert str(self.opening_tag.name) == self.closing_tag.name

    @property
    def name(self):
        return self.opening_tag.name

    @property
    def attributes(self):
        return self.opening_tag.attributes

    def __str__(self):
        if self.content is None:
            content_str = ""
        else:
            content_str = "".join(str(n) for n in self.content)
        return "".join(
            [str(self.opening_tag), content_str, str(self.closing_tag or "")]
        )


@attr.s(frozen=True)
class String(Node):
    value = attr.ib()  # str
    quote = attr.ib()  # '"' | "'" | None

    def __str__(self):
        if self.quote is None:
            return str(self.value)
        return self.quote + str(self.value) + self.quote


@attr.s(frozen=True)
class Integer(Node):
    value = attr.ib()  # int
    has_percent = attr.ib()  # bool

    def __str__(self):
        return str(self.value) + ("%" if self.has_percent else "")


@attr.s(frozen=True)
class Attribute(Node):
    name = attr.ib()  # str
    value = attr.ib()  # String | Integer

    def __str__(self):
        return "{}={}".format(self.name, self.value)


@attr.s(frozen=True)
class Comment(Node):
    text = attr.ib()  # str

    def __str__(self):
        return "<!--{}-->".format(self.text)


@attr.s(frozen=True)
class Jinja(Node):
    pass


@attr.s(frozen=True)
class JinjaVariable(Jinja):
    content = attr.ib()  # str
    left_plus = attr.ib(default=False)
    left_minus = attr.ib(default=False)
    right_minus = attr.ib(default=False)

    def __str__(self):
        return "".join(
            [
                "{{",
                "+" if self.left_plus else "",
                "-" if self.left_minus else "",
                " ",
                self.content,
                " ",
                "-" if self.right_minus else "",
                "}}",
            ]
        )


@attr.s(frozen=True)
class JinjaComment(Jinja):
    text = attr.ib()  # str

    def __str__(self):
        return "{##}"


@attr.s(frozen=True)
class JinjaTag(Jinja):
    name = attr.ib()
    content = attr.ib()  # str | None
    left_plus = attr.ib(default=False)
    left_minus = attr.ib(default=False)
    right_minus = attr.ib(default=False)

    def __str__(self):
        return "".join(
            [
                "{%",
                "+" if self.left_plus else "",
                "-" if self.left_minus else "",
                (" " + self.name) if self.name else "",
                (" " + self.content) if self.content else "",
                " ",
                "-" if self.right_minus else "",
                "%}",
            ]
        )


@attr.s(frozen=True)
class JinjaElementPart(Jinja):
    tag = attr.ib()  # JinjaTag
    content = attr.ib()  # Interpolated | None

    def __str__(self):
        return str(self.tag) + str(self.content or "")


@attr.s(frozen=True)
class JinjaElement(Jinja):
    parts = attr.ib()  # [JinjaElementPart]
    closing_tag = attr.ib()  # JinjaTag

    def __str__(self):
        return "".join(str(p) for p in self.parts) + str(self.closing_tag or "")


@attr.s(frozen=True)
class JinjaOptionalContainer(Jinja):
    first_opening_if = attr.ib()  # JinjaTag
    opening_tag = attr.ib()  # OpeningTag
    first_closing_if = attr.ib()  # JinjaTag
    content = attr.ib()  # Interpolated
    second_opening_if = attr.ib()  # JinjaTag
    closing_tag = attr.ib()  # ClosingTag
    second_closing_if = attr.ib()  # JinjaTag

    def __str__(self):
        nodes = [
            self.first_opening_if,
            self.opening_tag,
            self.first_closing_if,
            self.content,
            self.second_opening_if,
            self.closing_tag,
            self.second_closing_if,
        ]
        return "".join(str(n) for n in nodes)


@attr.s(frozen=True)
class InterpolatedBase(Node):
    nodes = attr.ib()  # [any | Jinja]

    @property
    def single_node(self):
        return self.nodes[0] if len(self.nodes) == 1 else None

    @property
    def single_str(self):
        node = self.single_node
        return node if isinstance(node, str) else None

    def __getitem__(self, index):
        return self.nodes.__getitem__(index)

    def __iter__(self):
        return self.nodes.__iter__()

    def __len__(self):
        return len(self.nodes)

    def __str__(self):
        return "".join(str(n) for n in self.nodes)


def _concat_strings(nodes):
    if len(nodes) <= 1:
        return nodes
    a, b, *rest = nodes
    if isinstance(a, str) and isinstance(b, str):
        return _concat_strings([a + b] + rest)
    return [a] + _concat_strings([b] + rest)


def _normalize_nodes(thing):
    if not isinstance(thing, list):
        nodes = [thing]
    else:
        nodes = thing
    return _concat_strings(nodes)


class Interpolated(InterpolatedBase):
    def __init__(self, *args, **kwargs):
        if len(args) == 1:
            assert "nodes" not in kwargs
            nodes = _normalize_nodes(args[0])
            super().__init__(nodes=nodes, **kwargs)
            return

        assert len(args) == 0
        kwargs = kwargs.copy()
        kwargs["nodes"] = _normalize_nodes(kwargs["nodes"])
        super().__init__(**kwargs)



from curlylint import ast
from curlylint.check_node import CheckNode, build_tree
from curlylint.issue import Issue

HTML_HAS_LANG = "html_has_lang"

RULE = {
    "id": "html_has_lang",
    "type": "accessibility",
    "docs": {
        "description": "`<html>` elements must have a `lang` attribute, using a [BCP 47](https://www.ietf.org/rfc/bcp/bcp47.txt) language tag.",
        "url": "https://www.curlylint.org/docs/rules/html_has_lang",
        "impact": "Serious",
        "tags": ["cat.language", "wcag2a", "wcag311"],
        "resources": [
            "[WCAG2.1 SC 3.1.1: Language of Page (Level A)](https://www.w3.org/WAI/WCAG21/Understanding/language-of-page)",
            "[axe-core, html-has-lang](https://dequeuniversity.com/rules/axe/3.5/html-has-lang)",
            "[axe-core, html-lang-valid](https://dequeuniversity.com/rules/axe/3.5/html-lang-valid)",
            "[eslint-plugin-jsx-a11y, html-has-lang](https://github.com/evcohen/eslint-plugin-jsx-a11y/blob/master/docs/rules/html-has-lang.md)",
        ],
    },
    "schema": {
        "$schema": "http://json-schema.org/draft/2019-09/schema#",
        "oneOf": [
            {
                "const": True,
                "title": "The `lang` attribute must be present.",
                "examples": [True],
            },
            {
                "type": "string",
                "title": "The `lang` attribute must match the configured language tag.",
                "examples": ["en-US"],
            },
            {
                "type": "array",
                "items": {"type": "string"},
                "uniqueItems": True,
                "title": "The `lang` attribute must match one of the configured language tags.",
                "examples": [["en", "en-US"]],
            },
        ],
    },
}


def find_valid(node, file, target_lang):
    name = getattr(node.value, "name", None)
    is_html = (
        isinstance(node.value, ast.Element) and name and name.lower() == "html"
    )

    if is_html:
        attributes = []
        if getattr(node.value, "opening_tag", None):
            attributes = {}
            for n in node.value.opening_tag.attributes.nodes:
                attributes[str(getattr(n, "name", ""))] = str(
                    getattr(n, "value", "")
                ).strip("\"'")
        if len(attributes) == 0 or "lang" not in attributes:
            return [
                Issue.from_node(
                    file,
                    node,
                    "The `<html>` tag should have a `lang` attribute with a valid value, describing the main language of the page",
                    "html_has_lang",
                )
            ]

        if target_lang[0] is True or attributes["lang"] in target_lang:
            return []
        else:
            return [
                Issue.from_node(
                    file,
                    node,
                    f"The `<html>` tag should have a `lang` attribute with a valid value, describing the main language of the page. Allowed values: {', '.join(target_lang)}",
                    "html_has_lang",
                )
            ]

    if not node.children:
        return []

    return sum(
        (find_valid(child, file, target_lang) for child in node.children), []
    )


def html_has_lang(file, target_lang):
    root = CheckNode(None)
    build_tree(root, file.tree)
    src = file.source.lower()

    target = (
        target_lang
        if isinstance(target_lang, (tuple, list))
        else (target_lang,)
    )

    if r"<html" in src:
        return find_valid(root, file, target)

    return []



from curlylint import ast
from curlylint.check_node import CheckNode, build_tree
from curlylint.issue import Issue

DJANGO_FORMS_RENDERING = "django_forms_rendering"

RULE = {
    "id": "django_forms_rendering",
    "type": "accessibility",
    "docs": {
        "description": "Disallows using Django’s convenience form rendering helpers, for which the markup isn’t screen-reader-friendly",
        "url": "https://www.curlylint.org/docs/rules/django_forms_rendering",
        "impact": "Serious",
        "tags": ["wcag2a", "wcag131"],
        "resources": [
            "[WCAG2.1 SC 1.3.1: Info and Relationships (Level A)](https://www.w3.org/WAI/WCAG21/Understanding/info-and-relationships.html)",
        ],
    },
    "schema": {
        "$schema": "http://json-schema.org/draft/2019-09/schema#",
        "oneOf": [
            {
                "const": True,
                "title": "Forms cannot be rendered with as_table, as_ul, or as_p",
                "examples": [True],
            },
            {"const": "as_p", "title": "Allows as_p", "examples": ["as_p"]},
        ],
    },
}


def find_valid(node, file, disallowed_variants):
    if isinstance(node.value, ast.JinjaVariable):
        matches = [v for v in disallowed_variants if v in node.value.content]
        if matches:
            return [
                Issue.from_node(
                    file,
                    node,
                    f"Avoid using `{matches[0]}` to render Django forms",
                    DJANGO_FORMS_RENDERING,
                )
            ]

    if not node.children:
        return []

    return sum(
        (
            find_valid(child, file, disallowed_variants)
            for child in node.children
        ),
        [],
    )


def django_forms_rendering(file, target):
    root = CheckNode(None)
    build_tree(root, file.tree)
    src = file.source.lower()

    if r".as_table" in src or ".as_ul" in src or ".as_p" in src:
        disallowed_variants = (
            ["as_table", "as_ul"]
            if target == "as_p"
            else ["as_table", "as_ul", "as_p"]
        )

        return find_valid(root, file, disallowed_variants)

    return []



import re

from curlylint import ast
from curlylint.issue import Issue, IssueLocation

WHITESPACE_INDENT_RE = re.compile(r"^\s*")
INDENT_RE = re.compile("^ *")

INDENT = "indent"

RULE = {
    "id": "indent",
    "type": "layout",
    "docs": {
        "description": "Enforce consistent indentation",
        "url": "https://www.curlylint.org/docs/rules/indent",
        "impact": "Developers only",
        "tags": [],
        "resources": [
            "Known issue: [Allow indentation inside template tags](https://github.com/thibaudcolas/curlylint/issues/6)",
        ],
    },
    "schema": {
        "$schema": "http://json-schema.org/draft/2019-09/schema#",
        "oneOf": [
            {"const": "tab", "title": "Use tabs.", "examples": ["tab"]},
            {
                "type": "integer",
                "minimum": 0,
                "title": "Use a given number of spaces.",
                "examples": [4],
            },
        ],
    },
}


def get_line_beginning(source, node):
    source = source[: node.begin.index]
    return source.split("\n")[-1]


def get_indent_level(source, node):
    """
    Returns the number of whitespace characters before the given node,
    in the first line of node.
    Returns `None` if some characters before the given node in this
    line aren’t whitespace.

    For example, if the source file contains `   <br /> ` on a line,
    `get_indent_level` will return 3 if called with the `<br />` tag
    as `node`.
    """
    beginning = get_line_beginning(source, node)
    if beginning and not beginning.isspace():
        return None
    return len(beginning)


def contains_exclusively(string, char):
    return string.replace(char, "") == ""


def truncate(s, length=16):
    return s[:length] + (s[length:] and "…")


def check_indentation(file, options):
    indent_size = options

    issues = []

    def add_issue(location, msg):
        issues.append(
            Issue.from_ast(file, location.line, location.column, msg, INDENT)
        )

    def check_indent(expected_level, node, inline=False, allow_same_line=False):
        node_level = get_indent_level(file.source, node)
        if node_level is None:
            if not inline and not allow_same_line:
                node_s = repr(truncate(str(node)))
                add_issue(node.begin, node_s + " should be on the next line")
            return

        if node_level != expected_level:
            msg = "Bad indentation, expected {}, got {}".format(
                expected_level, node_level
            )
            add_issue(node.begin, msg)

    def check_attribute(expected_level, attr, inline=False, **_):
        if not attr.value:
            return

        if attr.begin.line != attr.value.begin.line:
            add_issue(
                attr.begin,
                "The value must begin on line {}".format(attr.begin.line),
            )
        check_content(
            expected_level,
            attr.value,
            inline=attr.value.begin.line == attr.value.end.line,
            allow_same_line=True,
        )

    def check_opening_tag(expected_level, tag, inline=False, **_):
        if len(tag.attributes) and tag.begin.line != tag.end.line:
            first = tag.attributes[0]
            check_node(
                expected_level + indent_size,
                first,
                inline=isinstance(first, ast.Attribute),
            )
            attr_level = len(get_line_beginning(file.source, first))
            for attr in tag.attributes[1:]:
                # attr may be a JinjaElement
                check_node(
                    expected_level if inline else attr_level,
                    attr,
                    inline=isinstance(attr, ast.Attribute),
                )

    def check_comment(expected_level, tag, **_):
        pass

    def check_jinja_comment(expected_level, tag, **_):
        pass

    def check_jinja_tag(expected_level, tag, **_):
        pass

    def check_string(
        expected_level, string, inline=False, allow_same_line=False
    ):
        if string.value.begin.line != string.value.end.line:
            inline = False
        check_content(
            string.value.begin.column,
            string.value,
            inline=inline,
            allow_same_line=allow_same_line,
        )

    def check_integer(expected_level, integer, **_):
        pass

    def get_first_child_node(parent):
        for c in parent:
            if isinstance(c, ast.Node):
                return c
        return None

    def check_jinja_element_part(
        expected_level, part, inline=False, allow_same_line=False
    ):
        check_node(
            expected_level,
            part.tag,
            inline=inline,
            allow_same_line=allow_same_line,
        )
        if part.begin.line != part.end.line:
            inline = False
        shift = 0 if inline else indent_size
        content_level = expected_level + shift
        if part.content is not None:
            check_content(content_level, part.content, inline=inline)

    def check_jinja_optional_container_if(
        expected_level, o_if, html_tag, c_if, inline=False
    ):
        check_indent(expected_level, o_if, inline=inline)
        shift = 0 if inline else indent_size
        if isinstance(html_tag, ast.OpeningTag):
            check_opening_tag(expected_level + shift, html_tag, inline=inline)
        elif isinstance(html_tag, ast.ClosingTag):
            check_indent(expected_level + shift, html_tag, inline=inline)
        else:
            raise AssertionError("invalid tag")
        check_indent(expected_level, c_if, inline=inline)
        return inline

    def check_jinja_optional_container(
        expected_level, element, inline=False, **_
    ):
        if (
            element.first_opening_if.begin.line
            == element.second_opening_if.end.line
        ):
            inline = True

        inline = check_jinja_optional_container_if(
            expected_level,
            element.first_opening_if,
            element.opening_tag,
            element.first_closing_if,
            inline=inline,
        )

        check_content(expected_level, element.content, inline=inline)

        check_jinja_optional_container_if(
            expected_level,
            element.second_opening_if,
            element.closing_tag,
            element.second_closing_if,
            inline=inline,
        )

    def check_jinja_element(
        expected_level, element, inline=False, allow_same_line=False
    ):
        if element.begin.line == element.end.line:
            inline = True
        for part in element.parts:
            check_node(
                expected_level,
                part,
                inline=inline,
                allow_same_line=allow_same_line,
            )
        if element.closing_tag is not None:
            check_indent(expected_level, element.closing_tag, inline=inline)

    def check_jinja_variable(expected_level, var, **_):
        pass

    def check_element(expected_level, element, inline=False, **_):
        opening_tag = element.opening_tag
        closing_tag = element.closing_tag
        check_opening_tag(expected_level, opening_tag, inline=inline)
        if not closing_tag:
            return
        if inline or opening_tag.end.line == closing_tag.begin.line:
            check_content(expected_level, element.content, inline=True)
        else:
            check_content(expected_level + indent_size, element.content)
            check_indent(expected_level, closing_tag)

    def check_node(
        expected_level, node, inline=False, allow_same_line=False, **_
    ):
        check_indent(
            expected_level, node, inline=inline, allow_same_line=allow_same_line
        )

        types_to_functions = {
            ast.Attribute: check_attribute,
            ast.Comment: check_comment,
            ast.Element: check_element,
            ast.Integer: check_integer,
            ast.JinjaComment: check_jinja_comment,
            ast.JinjaElement: check_jinja_element,
            ast.JinjaElementPart: check_jinja_element_part,
            ast.JinjaOptionalContainer: check_jinja_optional_container,
            ast.JinjaTag: check_jinja_tag,
            ast.JinjaVariable: check_jinja_variable,
            ast.String: check_string,
        }

        func = types_to_functions.get(type(node))
        if func is None:
            raise Exception(
                "Unexpected {!r} node at {}".format(type(node), node.begin)
            )

        func(
            expected_level, node, inline=inline, allow_same_line=allow_same_line
        )

    def check_content_str(expected_level, string, parent_node):
        lines = string.split("\n")
        expected_indent = expected_level * " "

        indent = INDENT_RE.match(lines[0]).group(0)

        if len(indent) > 1:
            msg = (
                "Expected at most one space at the beginning of the text "
                "node, got {} spaces"
            ).format(len(indent))
            add_issue(parent_node.begin, msg)

        # skip the first line since there is certainly an HTML tag before
        for line in lines[1:]:
            if line.strip() == "":
                continue
            indent = INDENT_RE.match(line).group(0)
            if indent != expected_indent:
                msg = "Bad text indentation, expected {}, got {}".format(
                    expected_level, len(indent)
                )
                add_issue(parent_node.begin, msg)

    def check_content(
        expected_level, parent_node, inline=False, allow_same_line=False
    ):
        inline_parent = inline
        for i, child in enumerate(parent_node):
            next_child = get_first_child_node(parent_node[i + 1 :])

            if isinstance(child, str):
                check_content_str(expected_level, child, parent_node)
                if not child.strip(" "):
                    inline = True
                elif child.strip() and child.count("\n") <= 1:
                    inline = True
                elif (
                    next_child
                    and child.strip()
                    and not child.replace(" ", "").endswith("\n")
                ):
                    inline = True
                elif child.replace(" ", "").endswith("\n\n"):
                    inline = False
                if inline_parent and not inline:
                    msg = (
                        "An inline parent element must only contain "
                        "inline children"
                    )
                    add_issue(parent_node.begin, msg)
                continue

            if isinstance(child, ast.Node):
                if next_child and child.begin.line == next_child.end.line:
                    inline = True
                check_node(
                    expected_level,
                    child,
                    inline=inline,
                    allow_same_line=allow_same_line,
                )
                continue

            raise Exception()

    check_content(0, file.tree)

    return issues


def check_space_only_indent(file):
    issues = []
    for i, line in enumerate(file.lines):
        indent = WHITESPACE_INDENT_RE.match(line).group(0)
        if not contains_exclusively(indent, " "):
            loc = IssueLocation(file_path=file.path, line=i, column=0)
            issue = Issue(loc, "Should be indented with spaces", INDENT)
            issues.append(issue)
    return issues


def check_tab_only_indent(file):
    issues = []
    for i, line in enumerate(file.lines):
        indent = WHITESPACE_INDENT_RE.match(line).group(0)
        if not contains_exclusively(indent, "\t"):
            loc = IssueLocation(file_path=file.path, line=i, column=0)
            issue = Issue(loc, "Should be indented with tabs", INDENT)
            issues.append(issue)
    return issues


def indent(file, options):
    if options == "tab":
        return check_tab_only_indent(file)

    return check_space_only_indent(file) + check_indentation(file, options)



import json

import click


class RuleParamType(click.ParamType):
    """Validates and converts CLI-provided rule configuration."""

    name = "rule"

    def convert(self, value, param, ctx):
        try:
            rule_parts = value.split(":")
            code = rule_parts[0]
            options = json.loads(":".join(rule_parts[1:]))

            return {code: options}
        except TypeError:
            self.fail(
                "expected string rule, got "
                f"{value!r} of type {type(value).__name__}",
                param,
                ctx,
            )
        except ValueError:
            self.fail(
                f"{value!r} is not a valid rule configuration", param, ctx
            )


RULE = RuleParamType()



import unittest
import pytest

import parsy as P

from . import ast
from .ast import Location
from .parse import (
    comment,
    jinja_comment,
    make_attribute_parser,
    make_attribute_value_parser,
    make_attributes_parser,
    make_closing_tag_parser,
    make_opening_tag_parser,
    make_parser,
    tag_name,
    tag_name_char,
)

parser = make_parser()
element = parser["element"]
jinja = parser["jinja"]
content = parser["content"]
attribute_value = make_attribute_value_parser(jinja=jinja)
attribute = make_attribute_parser(jinja=jinja)
opening_tag = make_opening_tag_parser({}, jinja=jinja)


class DummyLocation:
    """
    Any instance of this class is equal to any Location.
    """

    def __eq__(self, other):
        if not isinstance(other, Location):
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return self.__class__.__name__ + "()"


def with_dummy_locations(node_class):
    def create_node(*args, **kwargs):
        kwargs = kwargs.copy()
        kwargs["begin"] = DummyLocation()
        kwargs["end"] = DummyLocation()

        try:
            return node_class(*args, **kwargs)
        except Exception as error:
            print(node_class)
            raise error

    return create_node


Attribute = with_dummy_locations(ast.Attribute)
Element = with_dummy_locations(ast.Element)
ClosingTag = with_dummy_locations(ast.ClosingTag)
Comment = with_dummy_locations(ast.Comment)
JinjaComment = with_dummy_locations(ast.JinjaComment)
Integer = with_dummy_locations(ast.Integer)
Interp = with_dummy_locations(ast.Interpolated)
JinjaComment = with_dummy_locations(ast.JinjaComment)
JinjaElement = with_dummy_locations(ast.JinjaElement)
JinjaElementPart = with_dummy_locations(ast.JinjaElementPart)
JinjaTag = with_dummy_locations(ast.JinjaTag)
JinjaVariable = with_dummy_locations(ast.JinjaVariable)
OpeningTag = with_dummy_locations(ast.OpeningTag)
String = with_dummy_locations(ast.String)


class TestUtil(unittest.TestCase):
    def test_dummy_location(self):
        dummy = DummyLocation()
        real = Location(0, 0, 0)
        self.assertEqual(dummy, real)
        self.assertEqual(real, dummy)

        self.assertEqual(dummy != real, False)  # lgtm [py/redundant-comparison]
        self.assertEqual(real != dummy, False)  # lgtm [py/redundant-comparison]


class TestParser(unittest.TestCase):
    def test_tag_name(self):
        self.assertEqual(tag_name_char.parse("a"), "a")
        self.assertEqual(tag_name.parse("bcd-ef9"), "bcd-ef9")
        self.assertEqual(tag_name.parse("clipPath"), "clipPath")
        self.assertEqual(tag_name.parse("HTML"), "HTML")

    def test_attribute_value(self):
        self.assertEqual(
            attribute_value.parse("hello-world"),
            String(value=Interp("hello-world"), quote=None),
        )

        self.assertEqual(
            attribute_value.parse("hello{{a}}world"),
            String(
                value=Interp(["hello", JinjaVariable(content="a"), "world"]),
                quote=None,
            ),
        )

        self.assertEqual(
            attribute_value.parse("123"), Integer(value=123, has_percent=False)
        )

        self.assertEqual(
            attribute_value.parse('"hello"'),
            String(value=Interp("hello"), quote='"'),
        )

        self.assertEqual(
            attribute_value.parse("'hello'"),
            String(value=Interp("hello"), quote="'"),
        )

        self.assertEqual(
            attribute_value.parse("''"), String(value=Interp([]), quote="'")
        )

        self.assertEqual(
            attribute_value.parse("'hello{{b}}world'"),
            String(
                value=Interp(["hello", JinjaVariable(content="b"), "world"]),
                quote="'",
            ),
        )

    def test_attribute(self):
        self.assertEqual(
            attribute.parse("hello=world"),
            Attribute(
                name=Interp("hello"),
                value=Interp(String(value=Interp("world"), quote=None)),
            ),
        )

        self.assertEqual(
            attribute.parse('a= "b"'),
            Attribute(
                name=Interp("a"),
                value=Interp(String(value=Interp("b"), quote='"')),
            ),
        )

        self.assertEqual(
            attribute.parse('A="b"'),
            Attribute(
                name=Interp("A"),
                value=Interp(String(value=Interp("b"), quote='"')),
            ),
        )

        self.assertEqual(
            attribute.parse('viewBox="b"'),
            Attribute(
                name=Interp("viewBox"),
                value=Interp(String(value=Interp("b"), quote='"')),
            ),
        )

        self.assertEqual(
            attribute.parse("a =b_c23"),
            Attribute(
                name=Interp("a"),
                value=Interp(String(value=Interp("b_c23"), quote=None)),
            ),
        )

        self.assertEqual(
            attribute.parse("valueless-attribute"),
            Attribute(name=Interp("valueless-attribute"), value=None),
        )

    def test_comment(self):
        self.assertEqual(
            comment.parse("<!--hello--world-->"), Comment(text="hello--world")
        )

    def test_jinja_comment(self):
        self.assertEqual(
            jinja_comment.parse("{# hello world #}"),
            JinjaComment(text="hello world"),
        )

    def test_opening_tag(self):
        self.assertEqual(
            opening_tag.parse("<div>"),
            OpeningTag(name="div", attributes=Interp([])),
        )

        self.assertEqual(
            opening_tag.parse("<div\n >"),
            OpeningTag(name="div", attributes=Interp([])),
        )

        self.assertEqual(
            opening_tag.parse('<div class="red" style="" >'),
            OpeningTag(
                name="div",
                attributes=Interp(
                    [
                        Attribute(
                            name=Interp("class"),
                            value=Interp(
                                String(value=Interp("red"), quote='"')
                            ),
                        ),
                        Attribute(
                            name=Interp("style"),
                            value=Interp(String(value=Interp([]), quote='"')),
                        ),
                    ]
                ),
            ),
        )

    def test_opening_tag_attributes_no_space(self):
        # See https://html.spec.whatwg.org/multipage/syntax.html#start-tags
        # "Attributes must be separated from each other by one or more ASCII whitespace."
        # See https://github.com/thibaudcolas/curlylint/issues/23#issuecomment-700622837
        with pytest.raises(
            P.ParseError, match="space\\(s\\) between attributes",
        ):
            opening_tag.parse(
                '<a class="govuk-button"href="/cookies">Set cookie preferences</a>'
            )

    def test_opening_tag_jinja_block_attributes_whitespace_after(self):
        opening_tag.parse(
            '<form{% if has_file_field %} enctype="multipart/form-data"{% endif %} action="{{ form_url }}">'
        )

    @pytest.mark.skip(reason="no way of currently testing this")
    def test_opening_tag_jinja_block_attributes_whitespace_before_37(self):
        # https://github.com/thibaudcolas/curlylint/issues/37
        opening_tag.parse(
            '<form {% if has_file_field %}enctype="multipart/form-data" {% endif %}action="{{ form_url }}">'
        )

    def test_closing_tag(self):
        closing_tag = make_closing_tag_parser(P.string("div"))
        self.assertEqual(closing_tag.parse("</div>"), ClosingTag(name="div"))

    def test_raw_text_elements(self):
        self.assertEqual(
            element.parse("<style a=b> <wont-be-parsed> </style>"),
            Element(
                content=" <wont-be-parsed> ",
                opening_tag=OpeningTag(
                    name="style",
                    attributes=Interp(
                        [
                            Attribute(
                                name=Interp("a"),
                                value=Interp(
                                    String(value=Interp("b"), quote=None)
                                ),
                            )
                        ]
                    ),
                ),
                closing_tag=ClosingTag(name="style"),
            ),
        )

    def test_element(self):
        self.assertEqual(
            element.parse("<div> hey </div>"),
            Element(
                opening_tag=OpeningTag(name="div", attributes=Interp([])),
                content=Interp([" hey "]),
                closing_tag=ClosingTag(name="div"),
            ),
        )

        attributes = [
            Attribute(
                name=Interp("onclick"),
                value=Interp(String(value=Interp([]), quote='"')),
            ),
            JinjaVariable(content="var"),
            Attribute(
                name=Interp("class"),
                value=Interp(String(value=Interp("red"), quote='"')),
            ),
        ]

        self.assertEqual(
            element.parse('<br onclick="" {{var}} class="red">'),
            Element(
                opening_tag=OpeningTag(
                    name="br", attributes=Interp(attributes)
                ),
                closing_tag=None,
                content=None,
            ),
        )

        src = "<{% if a %}bcd{% endif %}></{% if a %}bcd{% endif %}>"
        self.assertEqual(src, str(element.parse(src)))

        src = '<div{{ "ider" }}></div>'
        self.assertEqual('<div {{ "ider" }}></div>', str(element.parse(src)))

        src = '<div{% if a %}foo="bar"{% endif %}></div>'
        self.assertEqual(
            '<div {% if a %}foo="bar"{% endif %}></div>',
            str(element.parse(src)),
        )

        src = '<div{% if a %} foo="bar"  a=2 {% endif %}></div>'
        self.assertEqual(
            '<div {% if a %}foo="bar"a=2{% endif %}></div>',
            str(element.parse(src)),
        )

        src = "<colgroup></colgroup>"
        self.assertEqual(src, str(element.parse(src)))

    def test_self_closing_elements(self):
        self.assertEqual(
            element.parse("<br>"),
            Element(
                opening_tag=OpeningTag(name="br", attributes=Interp([])),
                content=None,
                closing_tag=None,
            ),
        )

        src = "<br />"
        self.assertEqual(src, str(element.parse(src)))

    def test_jinja_blocks(self):
        self.assertEqual(
            jinja.parse("{% name something == 123 %}"),
            JinjaElement(
                parts=[
                    JinjaElementPart(
                        tag=JinjaTag(name="name", content="something == 123"),
                        content=None,
                    )
                ],
                closing_tag=None,
            ),
        )

        self.assertEqual(
            jinja.parse("{% if a %}b{% else %}c{% endif %}"),
            JinjaElement(
                parts=[
                    JinjaElementPart(
                        tag=JinjaTag(name="if", content="a"),
                        content=Interp(["b"]),
                    ),
                    JinjaElementPart(
                        tag=JinjaTag(name="else", content=""),
                        content=Interp(["c"]),
                    ),
                ],
                closing_tag=JinjaTag(name="endif", content=""),
            ),
        )

        src = "{% if a %}b{% elif %}c{% elif %}d{% else %}e{% endif %}"
        self.assertEqual(src, str(jinja.parse(src)))

    def test_jinja_custom_tag_self_closing(self):
        self.assertEqual(
            jinja.parse("{% potato %}"),
            JinjaElement(
                parts=[
                    JinjaElementPart(
                        tag=JinjaTag(name="potato", content=""), content=None,
                    )
                ],
                closing_tag=None,
            ),
        )

    def test_jinja_custom_tag_open_close_unconfigured(self):
        with pytest.raises(P.ParseError):
            jinja.parse("{% of a %}c{% endof %}")

    def test_jinja_custom_tag_open_close_configured_deprecated(self):
        # Deprecated, will be removed in a future release.
        parser = make_parser({"jinja_custom_elements_names": [["of", "endof"]]})
        jinja = parser["jinja"]
        self.assertEqual(
            jinja.parse("{% of a %}c{% endof %}"),
            JinjaElement(
                parts=[
                    JinjaElementPart(
                        tag=JinjaTag(name="of", content="a"),
                        content=Interp(["c"]),
                    ),
                ],
                closing_tag=JinjaTag(name="endof", content=""),
            ),
        )

    def test_jinja_custom_tag_open_close_configured(self):
        parser = make_parser({"template_tags": [["of", "endof"]]})
        jinja = parser["jinja"]
        self.assertEqual(
            jinja.parse("{% of a %}c{% endof %}"),
            JinjaElement(
                parts=[
                    JinjaElementPart(
                        tag=JinjaTag(name="of", content="a"),
                        content=Interp(["c"]),
                    ),
                ],
                closing_tag=JinjaTag(name="endof", content=""),
            ),
        )

    def test_jinja_custom_tag_open_middle_close_unconfigured(self):
        with pytest.raises(P.ParseError):
            jinja.parse("{% of a %}b{% elseof %}c{% endof %}")

    def test_jinja_custom_tag_open_middle_close(self):
        parser = make_parser(
            {"jinja_custom_elements_names": [["of", "elseof", "endof"]]}
        )
        jinja = parser["jinja"]
        self.assertEqual(
            jinja.parse("{% of a %}b{% elseof %}c{% endof %}"),
            JinjaElement(
                parts=[
                    JinjaElementPart(
                        tag=JinjaTag(name="of", content="a"),
                        content=Interp(["b"]),
                    ),
                    JinjaElementPart(
                        tag=JinjaTag(name="elseof", content=""),
                        content=Interp(["c"]),
                    ),
                ],
                closing_tag=JinjaTag(name="endof", content=""),
            ),
        )

    def test_jinja_whitespace_controls(self):
        self.assertEqual(
            jinja.parse("{%- foo -%}"),
            JinjaElement(
                parts=[
                    JinjaElementPart(
                        tag=JinjaTag(
                            name="foo",
                            content="",
                            left_minus=True,
                            right_minus=True,
                        ),
                        content=None,
                    )
                ],
                closing_tag=None,
            ),
        )

        self.assertEqual(str(jinja.parse("{%- foo -%}")), "{%- foo -%}")
        self.assertEqual(str(jinja.parse("{%- foo %}")), "{%- foo %}")
        self.assertEqual(str(jinja.parse("{{- bar -}}")), "{{- bar -}}")
        self.assertEqual(str(jinja.parse("{{ bar -}}")), "{{ bar -}}")
        self.assertEqual(str(jinja.parse("{%+ foo %}")), "{%+ foo %}")
        self.assertEqual(str(jinja.parse("{{+ bar }}")), "{{+ bar }}")

    def test_doctype(self):
        self.assertEqual(
            content.parse("<!DOCTYPE html>"), Interp("<!DOCTYPE html>")
        )

    def test_attrs(self):
        attrs = make_attributes_parser({}, jinja)
        parse = attrs.parse

        self.assertEqual(
            str(parse("{% if %}{% endif %}")), "{% if %}{% endif %}"
        )
        self.assertEqual(
            str(parse("{% if %}  {% endif %}")), "{% if %}{% endif %}"
        )
        self.assertEqual(
            str(parse("{% if %}a=b{% endif %}")), "{% if %}a=b{% endif %}"
        )
        self.assertEqual(
            str(parse("{% if %} a=b {% endif %}")), "{% if %}a=b{% endif %}"
        )

    def test_optional_container(self):
        src = '{% if a %}<a href="b">{% endif %}c<b>d</b>{% if a %}</a>{% endif %}'
        self.assertEqual(src, str(content.parse(src)))

        src = """
        {% if a %} <a href="b"> {% endif %}
            c <b> d </b>
        {% if a %} </a> {% endif %}
        """
        content.parse(src)

    def test_whole_document(self):
        src = '<html lang="fr"><body>Hello<br></body></html>'
        self.assertEqual(src, str(element.parse(src)))



from functools import partial

import click

from curlylint.rules.aria_role.aria_role import aria_role
from curlylint.rules.django_forms_rendering.django_forms_rendering import (
    django_forms_rendering,
)
from curlylint.rules.html_has_lang.html_has_lang import html_has_lang
from curlylint.rules.image_alt.image_alt import image_alt
from curlylint.rules.indent.indent import indent
from curlylint.rules.meta_viewport.meta_viewport import meta_viewport
from curlylint.rules.no_autofocus.no_autofocus import no_autofocus
from curlylint.rules.tabindex_no_positive.tabindex_no_positive import (
    tabindex_no_positive,
)
from curlylint.util import flatten

err = partial(click.secho, fg="red", err=True)

checks = {
    "aria_role": aria_role,
    "django_forms_rendering": django_forms_rendering,
    "html_has_lang": html_has_lang,
    "image_alt": image_alt,
    "meta_viewport": meta_viewport,
    "no_autofocus": no_autofocus,
    "tabindex_no_positive": tabindex_no_positive,
    "indent": indent,
}


def check_rule(file, code: str, options):
    check = checks.get(code, None)

    if not check:
        err("Warning: rule `{}` does not exist.".format(code))
        return []

    return check(file, options)


def check_file(file, rules):
    return set(
        flatten(
            check_rule(file, code, options)
            for code, options in rules.items()
            if options != "off"
        )
    )


def check_files(files, rules):
    return flatten(check_file(file, rules) for file in files)



import re

# TODO: Move it elsewhere
import sys

import parsy as P

from .ast import (
    Attribute,
    ClosingTag,
    Comment,
    Element,
    Integer,
    Interpolated,
    Jinja,
    JinjaComment,
    JinjaElement,
    JinjaElementPart,
    JinjaOptionalContainer,
    JinjaTag,
    JinjaVariable,
    Location,
    OpeningTag,
    String,
)
from .util import flatten

sys.setrecursionlimit(10000)

# https://html.spec.whatwg.org/multipage/syntax.html#elements-2
# https://github.com/html5lib/html5lib-python/blob/0cae52b2073e3f2220db93a7650901f2200f2a13/html5lib/constants.py#L560
VOID_ELEMENTS = (
    "area",
    "base",
    "br",
    "col",
    "embed",
    "hr",
    "img",
    "input",
    "link",
    "meta",
    "param",
    "source",
    "track",
    "wbr",
)

# https://www.w3.org/TR/SVG2/eltindex.html
# https://developer.mozilla.org/en-US/docs/Web/SVG/Element
# Technically it’s valid for any SVG element to be written as a self-closing tag,
# for the parser this is equivalent to the element not having any content.
# In practice there are a lot of elements which wouldn’t be of much use if they didn’t
# have any content. This list omits those elements.
# This list also omits elements that have (next to) no browser implementation.
# Elements are only commented out so it’s easy to understand what isn’t there, and why.
SVG_SELF_CLOSING_ELEMENTS = (
    # "a",
    "animate",
    "animateMotion",
    "animateTransform",
    # "audio",
    # "canvas",
    "circle",
    # "clipPath",
    # "defs",
    # "desc",
    # "discard",
    "ellipse",
    "feBlend",
    "feColorMatrix",
    # "feComponentTransfer",
    "feComposite",
    "feConvolveMatrix",
    # "feDiffuseLighting",
    "feDisplacementMap",
    "feDistantLight",
    "feDropShadow",
    "feFlood",
    "feFuncA",
    "feFuncB",
    "feFuncG",
    "feFuncR",
    "feGaussianBlur",
    "feImage",
    # "feMerge",
    "feMergeNode",
    "feMorphology",
    "feOffset",
    "fePointLight",
    # "feSpecularLighting",
    "feSpotLight",
    "feTile",
    "feTurbulence",
    # "filter",
    # "foreignObject",
    # "g",
    # "hatchpath",
    # "iframe",
    "image",
    "line",
    # "linearGradient",
    # "marker",
    # "mask",
    # "metadata",
    "mpath",
    "path",
    # "pattern",
    "polygon",
    "polyline",
    # "radialGradient",
    "rect",
    # "script",
    "set",
    "stop",
    # "style",
    # "svg",
    # "switch",
    # "symbol",
    # "text",
    # "textPath",
    # "title",
    # "tspan",
    # "unknown",
    "use",
    # "video",
    # "view",
)

DEFAULT_JINJA_STRUCTURED_ELEMENTS_NAMES = [
    ("autoescape", "endautoescape"),
    ("block", "endblock"),
    ("blocktrans", "plural", "endblocktrans"),
    ("comment", "endcomment"),
    ("filter", "endfilter"),
    ("for", "else", "empty", "endfor"),
    ("if", "elif", "else", "endif"),
    ("ifchanged", "else", "endifchanged"),
    ("ifequal", "endifequal"),
    ("ifnotequal", "endifnotequal"),
    ("spaceless", "endspaceless"),
    ("verbatim", "endverbatim"),
    ("with", "endwith"),
]


# `\s` in an `str` pattern does match Unicode fancy spaces (especially the
# non-breaking ones). We’ll warn against some of these fancy spaces in the
# check phase.
# XXX: It could be better and simpler to only allow ASCII whitespaces here.
whitespace = P.regex(r"\s*")
mandatory_whitespace = P.regex(r"\s+")
spaces_between_attr = mandatory_whitespace.desc("space(s) between attributes")


def until(parser):
    return parser.should_fail(repr(parser)).then(P.any_char).many()


def combine_locations(begin_index, begin, result, end_index, end):
    begin_loc = Location(line=begin[0], column=begin[1], index=begin_index)
    end_loc = Location(line=end[0], column=end[1], index=end_index)
    locations = {"begin": begin_loc, "end": end_loc}

    return (locations, result)


def locate(parser):

    return P.seq(P.index, P.line_info, parser, P.index, P.line_info).combine(
        combine_locations
    )


def _combine_jinja_tag(locations, props):
    return JinjaTag(
        name=props["name"],
        content=props["extra_content"],
        left_plus=props["left_plus"],
        left_minus=props["left_minus"],
        right_minus=props["right_minus"],
        **locations,
    )


def _combine_jinja_variable(locations, props):
    return JinjaVariable(
        content=props["extra_content"],
        left_plus=props["left_plus"],
        left_minus=props["left_minus"],
        right_minus=props["right_minus"],
        **locations,
    )


def _combine_jinja_comment(locations, text):
    return JinjaComment(text=text, **locations)


def _combine_jinja_tag_like(locations, props):
    return (
        locations,
        {
            "left_plus": props[0] is not None,
            "left_minus": props[1] is not None,
            "name": props[2],
            "extra_content": props[3],
            "right_minus": props[4] is not None,
        },
    )


def make_jinja_tag_like_parser(name, ml="{", mr="}"):
    """
    Create parsers for Jinja variables and regular Jinja tags.

    `name` should be a parser to parse the tag name.
    """
    end = whitespace.then(P.string("-").optional()).skip(P.string(mr + "}"))
    return locate(
        P.seq(
            P.string("{" + ml).then(P.string("+").optional()),
            P.string("-").optional().skip(whitespace),
            name.skip(whitespace),
            until(end).concat(),
            end,
        )
    ).combine(_combine_jinja_tag_like)


jinja_variable = make_jinja_tag_like_parser(P.success(None), "{", "}").combine(
    _combine_jinja_variable
)


jinja_comment = locate(
    P.string("{#")
    .skip(whitespace)
    .then(until(whitespace + P.string("#}")).concat())
    .skip(whitespace + P.string("#}"))
).combine(_combine_jinja_comment)


def make_jinja_tag_parser(name_parser):
    return make_jinja_tag_like_parser(name_parser, "%", "%").combine(
        _combine_jinja_tag
    )


def _combine_jinja_element(locations, content):
    parts = list(flatten(content[0]))
    closing_tag = content[1] if len(content) == 2 else None

    e = JinjaElement(parts=parts, closing_tag=closing_tag, **locations)
    return e


def _combine_jinja_element_part(locations, props):
    tag, content = props
    return JinjaElementPart(tag=tag, content=content, **locations)


def make_jinja_element_part_parser(name_parser, content):
    return locate(P.seq(make_jinja_tag_parser(name_parser), content)).combine(
        _combine_jinja_element_part
    )


def make_jinja_element_parser(name_parsers, content):
    """
    `name_parsers` must be a list of tag name parsers. For example,
    `name_parsers` can be defined as follow in order to parse `if` statements:

        name_parsers = [P.string(n) for n in ['if', 'elif', 'else', 'endif']]
    """

    if len(name_parsers) == 1:
        tag = make_jinja_tag_parser(name_parsers[0])
        part = locate(P.seq(tag, P.success(None))).combine(
            _combine_jinja_element_part
        )
        parts = [part]
        end_tag_parser = None
    else:
        part_names = name_parsers[:-1]
        first_part = make_jinja_element_part_parser(
            part_names[0], content=content
        )
        next_parts = [
            make_jinja_element_part_parser(name, content=content).many()
            for name in part_names[1:]
        ]
        parts = [first_part] + next_parts
        end_tag_parser = make_jinja_tag_parser(name_parsers[-1])

    content = [P.seq(*parts)]
    if end_tag_parser:
        content.append(end_tag_parser)

    return locate(P.seq(*content)).combine(_combine_jinja_element)


jinja_name = (
    P.letter
    + (P.letter | P.decimal_digit | P.string("_"))  # type: ignore
    .many()
    .concat()
)


def interpolated(parser):
    def combine_interpolated(locations, result):
        return Interpolated(nodes=result, **locations)

    return locate(parser).combine(combine_interpolated)


tag_name_start_char = P.regex(r"[:a-zA-Z]")
tag_name_char = tag_name_start_char | P.regex(r"[0-9-_.]")
tag_name = tag_name_start_char + tag_name_char.many().concat()

dtd = P.regex(r"<![^>]*>")

# TODO This is overly restrictive on what can be inside attributes, it’s unclear why.
string_attribute_char = P.char_from("-_./+,?=:;#") | P.regex(r"[0-9a-zA-Z]")


def make_quoted_string_attribute_parser(quote, jinja):
    """
    quote: A single or a double quote
    """

    def combine(locations, value):
        return String(value=value, quote=quote, **locations)

    value_char = P.regex(r"[^<]", flags=re.DOTALL)
    value = interpolated(
        P.string(quote)
        .should_fail("no " + quote)
        .then(jinja | value_char)
        .many()
    )

    return locate(P.string(quote).then(value).skip(P.string(quote))).combine(
        combine
    )


def _combine_string_attribute_value(locations, value):
    return String(value=value, quote=None, **locations)


def _combine_int_attribute_value(locations, value):
    return Integer(value=value, has_percent=False, **locations)


int_attribute_value = locate(P.regex(r"[0-9]+").map(int)).combine(
    _combine_int_attribute_value
)


def _combine_attribute(locations, props):
    name, equal_and_value = props
    value = None if equal_and_value is None else equal_and_value["value"]
    return Attribute(name=name, value=value, **locations)


def make_attribute_value_parser(jinja):
    string_attribute_value = locate(
        interpolated((jinja | string_attribute_char).at_least(1))
    ).combine(_combine_string_attribute_value)

    return (
        make_quoted_string_attribute_parser('"', jinja)
        | make_quoted_string_attribute_parser("'", jinja)
        | int_attribute_value
        | string_attribute_value
    ).desc("attribute value")


attr_name_start_char = P.regex(r"[:a-zA-Z]")
attr_name_char = attr_name_start_char | P.regex(r"[0-9A-Z-_.]")
attr_name = attr_name_start_char + attr_name_char.many().concat()


def make_attribute_parser(jinja):
    attribute_value = make_attribute_value_parser(jinja)
    return (
        locate(
            P.seq(
                interpolated(attr_name),
                whitespace.then(
                    P.seq(
                        P.string("=").skip(whitespace).tag("equal"),
                        interpolated(attribute_value).tag("value"),
                    ).map(dict)
                ).optional(),
            )
        )
        .combine(_combine_attribute)
        .desc("attribute")
    )


def make_attributes_parser(config, jinja):
    attribute = make_attribute_parser(jinja)

    jinja_attr = make_jinja_parser(
        config,
        interpolated(
            whitespace.then((attribute | jinja).sep_by(whitespace)).skip(
                whitespace
            )
        ),
    )

    attrs = interpolated(
        (
            whitespace.then(jinja_attr) | spaces_between_attr.then(attribute)
        ).many()
    )

    return attrs


def _combine_comment(locations, text):
    return Comment(text=text, **locations)


comment = locate(
    P.string("<!--")
    .then(P.regex(r".*?(?=-->)", flags=re.DOTALL))
    .skip(P.string("-->"))
).combine(_combine_comment)


def _combine_opening_tag(locations, props):
    _lt, tag_name, attributes, slash, _gt = props
    return OpeningTag(
        name=tag_name, attributes=attributes, slash=slash, **locations
    )


def _combine_closing_tag(locations, name):
    return ClosingTag(name=name, **locations)


def make_closing_tag_parser(tag_name_parser):
    return locate(
        P.string("</").then(tag_name_parser).skip(P.string(">"))
    ).combine(_combine_closing_tag)


def _combine_slash(locations, _):
    return locations["begin"]


def make_opening_tag_parser(
    config, jinja, tag_name_parser=None, allow_slash=False, mandate_slash=False
):
    attributes = make_attributes_parser(config, jinja)

    if not tag_name_parser:
        tag_name_parser = tag_name | jinja

    if allow_slash:
        slash = (
            locate(P.string("/").skip(whitespace))
            .combine(_combine_slash)
            .optional()
        )
    elif mandate_slash:
        slash = locate(P.string("/").skip(whitespace)).combine(_combine_slash)
    else:
        slash = P.success(None)

    return locate(
        P.seq(
            P.string("<"),
            tag_name_parser,
            attributes.skip(whitespace),
            slash,
            P.string(">"),
        )
    ).combine(_combine_opening_tag)


def _combine_element(locations, props):
    opening_tag, content, closing_tag = props
    return Element(
        opening_tag=opening_tag,
        closing_tag=closing_tag,
        content=content,
        **locations,
    )


def make_raw_text_element_parser(config, tag_name, jinja):
    """
    Used for <style> and <script>.
    """
    opening_tag = make_opening_tag_parser(
        config, tag_name_parser=P.string(tag_name), jinja=jinja
    )

    body = P.regex(r".*?(?=</" + tag_name + ">)", flags=re.DOTALL)

    closing_tag = make_closing_tag_parser(P.string(tag_name))

    return locate(P.seq(opening_tag, body, closing_tag)).combine(
        _combine_element
    )


slow_text_char = (
    (P.string("<") | P.string("{%") | P.string("{#") | P.string("{{"))
    .should_fail("text")
    .then(P.any_char)
)

# Not as precise as slow_text_char but must faster
quick_text = P.regex(r"[^{<]+", flags=re.DOTALL)


def _combine_optional_container(locations, nodes):
    return JinjaOptionalContainer(
        first_opening_if=nodes[0],
        opening_tag=nodes[1],
        first_closing_if=nodes[2],
        content=nodes[3],
        second_opening_if=nodes[4],
        closing_tag=nodes[5],
        second_closing_if=nodes[6],
        **locations,
    )


# Awkward hack to handle optional HTML containers, for example:
#
#   {% if a %}
#     <div>
#   {% endif %}
#   foo
#   {% if a %}
#     </div>
#   {% endif %}
#
# Currently, this only works with `if` statements and the two conditions
# must be exactly the same.
def make_jinja_optional_container_parser(config, content, jinja):
    jinja_if = make_jinja_tag_parser(P.string("if"))
    jinja_endif = make_jinja_tag_parser(P.string("endif"))
    opening_tag = make_opening_tag_parser(config, jinja, allow_slash=False)

    @P.generate
    def opt_container_impl():
        o_first_if_node = yield jinja_if.skip(whitespace)
        o_tag_node = yield opening_tag.skip(whitespace)
        c_first_if_node = yield jinja_endif

        content_nodes = yield content

        o_second_if_node = yield jinja_if.skip(whitespace)
        if o_second_if_node.content != o_first_if_node.content:
            yield P.fail("expected `{% if " + content + " %}`")
            return
        html_tag_name = o_tag_node.name
        if isinstance(html_tag_name, str):
            closing_tag = make_closing_tag_parser(P.string(html_tag_name))
        else:
            assert isinstance(html_tag_name, Jinja)
            closing_tag = make_closing_tag_parser(jinja)
        c_tag_node = yield closing_tag.skip(whitespace)
        c_second_if_node = yield jinja_endif

        return [
            o_first_if_node,
            o_tag_node,
            c_first_if_node,
            content_nodes,
            o_second_if_node,
            c_tag_node,
            c_second_if_node,
        ]

    return locate(opt_container_impl).combine(_combine_optional_container)


def make_jinja_parser(config, content):
    # Allow to override elements with the configuration
    jinja_structured_elements_names = dict(
        (names[0], names)
        for names in (
            DEFAULT_JINJA_STRUCTURED_ELEMENTS_NAMES
            + config.get("template_tags", [])
            # Deprecated, will be removed in a future release.
            + config.get("jinja_custom_elements_names", [])
        )
    ).values()

    jinja_structured_element = P.alt(
        *[
            make_jinja_element_parser(
                [P.string(name) for name in names], content=content
            )
            for names in jinja_structured_elements_names
        ]
    )

    # These tag names can't begin a Jinja element
    jinja_intermediate_tag_names = set(
        n for _, *sublist in jinja_structured_elements_names for n in sublist
    )

    jinja_intermediate_tag_name = P.alt(
        *(P.string(n) for n in jinja_intermediate_tag_names)
    )

    jinja_element_single = make_jinja_element_parser(
        [
            P.alt(
                # HACK: If we allow `{% if %}`s without `{% endif %}`s here,
                # `make_jinja_optional_container_parser` doesn’t work. It
                # is probably better to reject any structured tag name here.
                P.string("if"),
                jinja_intermediate_tag_name,
            )
            .should_fail("not an intermediate Jinja tag name")
            .then(jinja_name)
        ],
        content=content,
    )

    jinja_element = jinja_structured_element | jinja_element_single

    return jinja_variable | jinja_comment | jinja_element


def make_container_element_parser(config, content, jinja):
    opening_tag = make_opening_tag_parser(config, jinja=jinja)

    @P.generate
    def container_element_impl():
        o_tag_node = yield opening_tag
        content_nodes = yield content
        tag_name = o_tag_node.name
        if isinstance(tag_name, str):
            closing_tag = make_closing_tag_parser(P.string(tag_name))
        else:
            assert isinstance(tag_name, Jinja)
            closing_tag = make_closing_tag_parser(jinja)
        c_tag_node = yield closing_tag
        return [o_tag_node, content_nodes, c_tag_node]

    return locate(container_element_impl).combine(_combine_element)


def make_element_parser(config, content, jinja):
    container_element = make_container_element_parser(
        config, content=content, jinja=jinja
    )

    void_element_opening_tag = make_opening_tag_parser(
        config,
        tag_name_parser=P.string_from(*VOID_ELEMENTS),
        allow_slash=True,
        jinja=jinja,
    )

    void_element = locate(
        P.seq(
            void_element_opening_tag.skip(whitespace),
            P.success(None),  # No content
            P.success(None),  # No closing tag
        )
    ).combine(_combine_element)

    svg_self_closing_tag = make_opening_tag_parser(
        config,
        tag_name_parser=P.string_from(*SVG_SELF_CLOSING_ELEMENTS),
        mandate_slash=True,
        jinja=jinja,
    )

    svg_self_closing_element = locate(
        P.seq(
            svg_self_closing_tag.skip(whitespace),
            P.success(None),  # No content
            P.success(None),  # No closing tag
        )
    ).combine(_combine_element)

    style = make_raw_text_element_parser(config, "style", jinja=jinja)
    script = make_raw_text_element_parser(config, "script", jinja=jinja)

    return (
        style
        | script
        | void_element
        | svg_self_closing_element
        | container_element
    )


def make_parser(config=None):
    if config is None:
        config = {}

    content_ = None

    @P.generate
    def content():
        c = yield content_
        return c

    jinja = make_jinja_parser(config, content)

    element = make_element_parser(config, content, jinja)

    opt_container = make_jinja_optional_container_parser(config, content, jinja)

    content_ = interpolated(
        (
            quick_text
            | comment
            | dtd
            | element
            | opt_container
            | jinja
            | slow_text_char
        ).many()
    )

    return {"content": content_, "jinja": jinja, "element": element}



import json

import click


class TemplateTagsParamType(click.ParamType):
    """
    Validates and converts CLI-provided template tags configuration.
    Expects: --template-tags '[["cache", "endcache"]]'
    """

    name = "template tags"

    def convert(self, value, param, ctx):
        try:
            if isinstance(value, str):
                template_tags = json.loads(value)
            else:
                template_tags = value
            # Validate the expected list of lists.
            if not isinstance(template_tags, (list, tuple)):
                raise ValueError
            for tags in template_tags:
                if not isinstance(tags, (list, tuple)):
                    raise ValueError
            return template_tags
        except ValueError:
            self.fail(
                f"expected a list of lists of tags as JSON, got {value!r}",
                param,
                ctx,
            )


TEMPLATE_TAGS = TemplateTagsParamType()



from dataclasses import dataclass  # NOQA
from functools import partial
from pathlib import Path
from typing import List

import click

from curlylint.formatters.compact import format_compact
from curlylint.formatters.json import format_json
from curlylint.formatters.stylish import format_stylish
from curlylint.issue import Issue

out = partial(click.secho, bold=True, err=True)
err = partial(click.secho, fg="red", err=True)

formatters = {
    "compact": format_compact,
    "json": format_json,
    "stylish": format_stylish,
}


@dataclass
class Report:
    """Taken from black:
    https://github.com/psf/black/blob/959848c17639bfc646128f6b582c5858164a5001/black.py#L3640
    Provides a reformatting counter. Can be rendered with `str(report)`."""

    quiet: bool = False
    verbose: bool = False
    failure_count: int = 0
    format: str = "compact"

    def path_ignored(self, path: Path, message: str) -> None:
        if self.verbose:
            out(f"{path} ignored: {message}", bold=False)

    def print_issues(self, issues: List[Issue]):
        self.failure_count += len(issues)
        formatter = formatters[self.format]

        formatter(issues)

    @property
    def return_code(self) -> int:
        """Return the exit code that the app should use."""
        if self.failure_count:
            return 1

        return 0

    def __str__(self) -> str:
        """Render a color report of the current state.
        Use `click.unstyle` to remove colors.
        """
        report = []

        if self.failure_count:
            s = "s" if self.failure_count > 1 else ""
            report.append(
                click.style(f"{self.failure_count} error{s} reported", fg="red")
            )

        return ", ".join(report)



from curlylint import ast
from curlylint.ast import JinjaElement
from curlylint.check_node import CheckNode, build_tree
from curlylint.issue import Issue

IMAGE_ALT = "image_alt"

RULE = {
    "id": "image_alt",
    "type": "accessibility",
    "docs": {
        "description": "`<img>` elements must have a `alt` attribute, either with meaningful text, or an empty string for decorative images",
        "url": "https://www.curlylint.org/docs/rules/image_alt",
        "impact": "Serious",
        "tags": ["cat.language", "wcag2a", "wcag111"],
        "resources": [
            "[WCAG2.1 SC 1.1.1: Non-text Content (Level A)](https://www.w3.org/WAI/WCAG21/Understanding/non-text-content.html)",
            "[axe-core, image-alt](https://dequeuniversity.com/rules/axe/3.5/image-alt)",
            "[eslint-plugin-jsx-a11y, alt-text](https://github.com/evcohen/eslint-plugin-jsx-a11y/blob/master/docs/rules/alt-text.md)",
        ],
    },
    "schema": {
        "$schema": "http://json-schema.org/draft/2019-09/schema#",
        "oneOf": [
            {
                "const": True,
                "title": "The `alt` attribute must be present.",
                "examples": [True],
            },
        ],
    },
}


def find_valid(node, file, target_alt):
    name = getattr(node.value, "name", None)
    is_img = (
        isinstance(node.value, ast.Element) and name and name.lower() == "img"
    )

    if is_img:
        attributes = []
        if getattr(node.value, "opening_tag", None):
            attributes = {}
            for n in node.value.opening_tag.attributes.nodes:
                if not isinstance(n, JinjaElement):
                    attributes[str(n.name)] = str(n.value).strip("\"'")

        if len(attributes) == 0 or "alt" not in attributes:
            return [
                Issue.from_node(
                    file,
                    node,
                    "The `<img>` tag must have a `alt` attribute, either with meaningful text, or an empty string for decorative images",
                    "image_alt",
                )
            ]

    if not node.children:
        return []

    return sum(
        (find_valid(child, file, target_alt) for child in node.children), []
    )


def image_alt(file, target_alt):
    root = CheckNode(None)
    build_tree(root, file.tree)
    src = file.source.lower()

    if r"<img" in src:
        return find_valid(root, file, target_alt)

    return []



__name__ = "curlylint"
__version__ = "0.13.0"
__description__ = "{{ 🎀}} Experimental HTML templates linting for Jinja, Nunjucks, Django templates, Twig, Liquid"
__author__ = "Thibaud Colas"
__author_email__ = "thibaudcolas@gmail.com"
__url__ = "https://github.com/thibaudcolas/curlylint"
__license__ = "MIT"
__copyright__ = "Copyright 2020-present Thibaud Colas"

if __name__ == "__main__":
    from .cli import patched_main

    patched_main()



import sys
from contextlib import contextmanager
from io import BytesIO, TextIOWrapper
from typing import (
    Any,
    BinaryIO,
    Generator,
)

from click.testing import CliRunner


class BlackRunner(CliRunner):
    """Borrowed from black.
    https://github.com/psf/black/blob/5446a92f0161e398de765bf9532d8c76c5652333/tests/test_black.py#L101
    Modify CliRunner so that stderr is not merged with stdout.
    This is a hack that can be removed once we depend on Click 7.x"""

    def __init__(self) -> None:
        self.stderrbuf = BytesIO()
        self.stdoutbuf = BytesIO()
        self.stdout_bytes = b""
        self.stderr_bytes = b""
        super().__init__()

    @contextmanager
    def isolation(
        self, *args: Any, **kwargs: Any
    ) -> Generator[BinaryIO, None, None]:
        with super().isolation(*args, **kwargs) as output:
            try:
                hold_stderr = sys.stderr
                sys.stderr = TextIOWrapper(
                    self.stderrbuf, encoding=self.charset
                )
                yield output
            finally:
                self.stdout_bytes = sys.stdout.buffer.getvalue()  # type: ignore
                self.stderr_bytes = sys.stderr.buffer.getvalue()  # type: ignore
                sys.stderr = hold_stderr



import json
from typing import Callable

from curlylint.issue import Issue
from curlylint.lint import parse_source


class RulesTestMeta(type):
    """
    Generates test cases dynamically.
    See http://stackoverflow.com/a/20870875/1798491
    """

    def __new__(mcs, name, bases, tests):
        def gen_test(item) -> Callable[[None], None]:
            def test(self):
                errors, file = parse_source("test.html", {}, item["template"])
                output = list(map(Issue.from_dict, item["output"]))
                self.assertEqual(tests["rule"](file, item["config"]), output)

            return test

        fixtures = json.loads(open(tests["fixtures"], "r").read())

        for item in fixtures:
            test_label = item["label"].lower().replace(" ", "_")
            test_name = f"test_{test_label}"
            tests[test_name] = gen_test(item)

        return type.__new__(mcs, name, bases, tests)



class CheckNode:
    def __init__(self, value):
        self.value = value
        self.children = []

    def __str__(self, level=0):
        name = getattr(self.value, "name", None)

        attributes = []
        if getattr(self.value, "opening_tag", None):
            attributes = [
                (
                    str(getattr(n, "name", "")),
                    str(getattr(n, "value", "")).strip("\"'"),
                )
                for n in self.value.opening_tag.attributes.nodes
            ]

        result = (
            "  " * level
            + "{}: name={!r} attributes={!r}".format(
                type(self.value), name, attributes
            )
            + "\n"
        )

        for child in self.children:
            result += child.__str__(level + 1)

        return result


def build_tree(root, node):
    if isinstance(node, str) or node is None:
        return

    for child in node.nodes:
        new_node = CheckNode(child)
        if getattr(child, "content", None):
            build_tree(new_node, child.content)
        root.children.append(new_node)



import json
from typing import List

from curlylint.issue import Issue


def format_json(issues: List[Issue]):
    sorted_issues = sorted(
        issues,
        key=lambda i: (
            i.location.file_path,
            i.location.line,
            i.location.column,
        ),
    )

    output = []

    for issue in sorted_issues:
        output.append(
            {
                "file_path": str(issue.location.file_path),
                "line": issue.location.line,
                "column": issue.location.column,
                "message": issue.message,
                "code": issue.code,
            }
        )

    print(json.dumps(output))



from typing import Any, Optional

def string(  # noqa: E704, E302
    expected_string: str, transform: Optional[Any] = None
) -> Any: ...  # noqa: E704
def regex(exp: str, flags: int = 0) -> Any: ...  # noqa: E704
def success(val: Any) -> Any: ...  # noqa: E704
def letter() -> Any: ...  # noqa: E704
def decimal_digit() -> Any: ...  # noqa: E704
def any_char() -> Any: ...  # noqa: E704
def char_from(characters: str) -> Any: ...  # noqa: E704, E302

class ParseError(RuntimeError):  # noqa: E704, E302
    def line_info(self) -> str: ...  # noqa: E704



from functools import lru_cache, partial
from pathlib import Path
from typing import Any, Dict, Iterable, Iterator, List, Optional, Pattern, Union

import click
import toml
from pathspec import PathSpec

from . import __version__
from .report import Report

out = partial(click.secho, bold=True, err=True)


@lru_cache()
def find_project_root(srcs: Iterable[str]) -> Path:
    """Taken from black:
    https://github.com/psf/black/blob/959848c17639bfc646128f6b582c5858164a5001/black.py#L237
    Return a directory containing .git, .hg, or pyproject.toml.
    That directory can be one of the directories passed in `srcs` or their
    common parent.
    If no directory in the tree contains a marker that would specify it's the
    project root, the root of the file system is returned.
    """
    if not srcs:
        return Path("/").resolve()

    common_base = min(Path(src).resolve() for src in srcs)
    if common_base.is_dir():
        # Append a fake file so `parents` below returns `common_base_dir`, too.
        common_base /= "fake-file"
    for directory in common_base.parents:
        if (directory / ".git").exists():
            return directory

        if (directory / ".hg").is_dir():
            return directory

        if (directory / "pyproject.toml").is_file():
            return directory

    return directory


def find_pyproject_toml(path_search_start: str) -> Optional[str]:
    """Find the absolute filepath to a pyproject.toml if it exists"""
    path_project_root = find_project_root(path_search_start)
    path_pyproject_toml = path_project_root / "pyproject.toml"
    return str(path_pyproject_toml) if path_pyproject_toml.is_file() else None


def parse_pyproject_toml(path_config: str) -> Dict[str, Any]:
    """Parse a pyproject toml file, pulling out relevant parts for the linter
    If parsing fails, will raise a toml.TomlDecodeError
    """
    pyproject_toml = toml.load(path_config)
    config = pyproject_toml.get("tool", {}).get("curlylint", {})
    return {k.replace("--", "").replace("-", "_"): v for k, v in config.items()}


def dump_toml_config(ctx: click.Context, config: Dict[str, Any] = None):
    """Prints the provided config object as TOML, if present."""
    if not config:
        out(
            "Oops! Something went wrong! :( curlylint couldn’t find a configuration file.\n"
            f"curlylint: v{__version__}.\n\n"
        )
        ctx.exit(2)

    print(toml.dumps({"tool": {"curlylint": config}}))
    ctx.exit(0)


def read_pyproject_toml(
    ctx: click.Context,
    param: click.Parameter,
    value: Union[str, int, bool, None],
) -> Optional[str]:
    """Inject configuration from "pyproject.toml" into defaults in `ctx`.
    Returns the path to a successfully found and read configuration file, None
    otherwise.
    """
    assert not isinstance(value, (int, bool)), "Invalid parameter type passed"
    print_config = ctx.params.get("print_config", False)

    if not value:
        value = find_pyproject_toml(ctx.params.get("src", ()))
        if value is None:
            if print_config:
                dump_toml_config(ctx)
            return None

    try:
        config = parse_pyproject_toml(value)
    except (toml.TomlDecodeError, OSError) as e:
        raise click.FileError(
            filename=value, hint=f"Error reading configuration file: {e}"
        )

    if not config:
        if print_config:
            dump_toml_config(ctx)
        return None

    if print_config:
        dump_toml_config(ctx, config)

    if ctx.default_map is None:
        ctx.default_map = {}
    ctx.default_map.update(config)  # type: ignore  # bad types in .pyi
    return value


@lru_cache()
def get_gitignore(root: Path) -> PathSpec:
    """ Return a PathSpec matching gitignore content if present."""
    gitignore = root / ".gitignore"
    lines: List[str] = []
    if gitignore.is_file():
        with gitignore.open() as gf:
            lines = gf.readlines()
    return PathSpec.from_lines("gitwildmatch", lines)


def gen_template_files_in_dir(
    path: Path,
    root: Path,
    include: Pattern[str],
    exclude: Pattern[str],
    report: "Report",
    gitignore: PathSpec,
) -> Iterator[Path]:
    """Generate all files under `path` whose paths are not excluded by the
    `exclude` regex, but are included by the `include` regex.
    Symbolic links pointing outside of the `root` directory are ignored.
    `report` is where output about exclusions goes.
    """
    assert (
        root.is_absolute()
    ), f"INTERNAL ERROR: `root` must be absolute but is {root}"
    for child in path.iterdir():
        # First ignore files matching .gitignore
        if gitignore.match_file(child.as_posix()):
            report.path_ignored(child, "matches the .gitignore file content")
            continue

        # Then ignore with `exclude` option.
        try:
            normalized_path = "/" + child.resolve().relative_to(root).as_posix()
        except OSError as e:
            report.path_ignored(child, f"cannot be read because {e}")
            continue

        except ValueError:
            if child.is_symlink():
                report.path_ignored(
                    child, f"is a symbolic link that points outside {root}"
                )
                continue

            raise

        if child.is_dir():
            normalized_path += "/"

        exclude_match = exclude.search(normalized_path)
        if exclude_match and exclude_match.group(0):
            report.path_ignored(
                child, "matches the --exclude regular expression"
            )
            continue

        if child.is_dir():
            yield from gen_template_files_in_dir(
                child, root, include, exclude, report, gitignore
            )

        elif child.is_file():
            include_match = include.search(normalized_path)
            if include_match:
                yield child



from curlylint import ast
from curlylint.check_node import CheckNode, build_tree
from curlylint.issue import Issue

NO_AUTOFOCUS = "no_autofocus"

RULE = {
    "id": "no_autofocus",
    "type": "accessibility",
    "docs": {
        "description": "Enforce autofocus is not used on inputs.  Autofocusing elements can cause usability issues for sighted and non-sighted users.",
        "url": "https://www.curlylint.org/docs/rules/no_autofocus",
        "impact": "Serious",
        "tags": ["cat.language", "wcag2a"],
        "resources": [
            "[WHATWG HTML Standard, The autofocus attribute](https://html.spec.whatwg.org/multipage/interaction.html#attr-fe-autofocus)",
            "[The accessibility of HTML 5 autofocus](https://www.brucelawson.co.uk/2009/the-accessibility-of-html-5-autofocus/)",
            "[MDN: input `autofocus` attribute usage considerations](https://developer.mozilla.org/en-US/docs/Web/HTML/Element/input#htmlattrdefautofocus)",
        ],
    },
    "schema": {
        "$schema": "http://json-schema.org/draft/2019-09/schema#",
        "oneOf": [
            {
                "const": True,
                "title": "The `autofocus` attribute must not be used.",
                "examples": [True],
            },
        ],
    },
}


def find_valid(node, file, target_alt):
    name = getattr(node.value, "name", None)
    is_input = (
        isinstance(node.value, ast.Element) and name and name.lower() == "input"
    )

    if is_input:
        attributes = []
        if getattr(node.value, "opening_tag", None):
            attributes = {}
            for n in node.value.opening_tag.attributes.nodes:
                attributes[str(n.name)] = str(n.value).strip("\"'")

        if "autofocus" in attributes:
            return [
                Issue.from_node(
                    file,
                    node,
                    "Do not use the `autofocus` attribute, which causes issues for screen reader users",
                    "no_autofocus",
                )
            ]

    if not node.children:
        return []

    return sum(
        (find_valid(child, file, target_alt) for child in node.children), []
    )


def no_autofocus(file, target_alt):
    root = CheckNode(None)
    build_tree(root, file.tree)
    src = file.source.lower()

    if r"autofocus" in src:
        return find_valid(root, file, target_alt)

    return []



from typing import Any, Iterable

class PathSpec:  # noqa: E302
    def match_file(self, file: str) -> bool: ...  # noqa: E704
    @classmethod
    def from_lines(  # noqa: E704
        cls, pattern_factory: str, lines: Iterable[str]
    ) -> Any: ...



import attr

from .file import File


@attr.s(frozen=True)
class IssueLocation:
    """Location inside a file: file path, and line/col offsets."""

    file_path = attr.ib()  # Path
    column = attr.ib()
    line = attr.ib()

    def __str__(self):
        return "{}:{}:{}".format(self.file_path, self.line + 1, self.column)

    @staticmethod
    def from_ast(file_path, line, column):
        if isinstance(file_path, File):
            file_path = file_path.path

        return IssueLocation(file_path=file_path, line=line, column=column,)


@attr.s(frozen=True)
class Issue:
    """A problem to report on a given location of a file."""

    location = attr.ib()
    message = attr.ib()
    code = attr.ib()

    def __str__(self):
        return f"{self.location}: {self.message} ({self.code})"

    def __attrs_post_init__(self):
        assert isinstance(self.location, IssueLocation)

    @staticmethod
    def from_ast(file, line, column, message, code):
        return Issue(IssueLocation.from_ast(file, line, column), message, code)

    @staticmethod
    def from_node(file, node, message, code):
        return Issue.from_ast(
            file.path,
            node.value.begin.line + 1,
            node.value.begin.column + 1,
            message,
            code,
        )

    @staticmethod
    def from_dict(issue):
        return Issue.from_ast(**issue)



import unittest

from .util import flatten


class TestUtil(unittest.TestCase):
    def test_flatten(self):
        self.assertEqual(list(flatten(())), [])
        self.assertEqual(list(flatten([])), [])
        self.assertEqual(list(flatten((1,))), [1])
        self.assertEqual(
            list(flatten([2, [], (), [3, [(4, 5), (6,)]]])), [2, 3, 4, 5, 6]
        )



from collections.abc import Iterable


def flatten(_list):
    """
    Deeply flattens an iterable.
    """
    for el in _list:
        if isinstance(el, Iterable) and not isinstance(el, (str, bytes)):
            yield from flatten(el)
        else:
            yield el



import unittest

from io import BytesIO
from typing import List

from curlylint.tests.utils import BlackRunner

from curlylint.cli import main


class TestCLI(unittest.TestCase):
    """
    Heavily inspired by Black’s CLI tests.
    See https://github.com/psf/black/blob/master/tests/test_black.py.
    """

    def invoke_curlylint(
        self, exit_code: int, args: List[str], input: str = None
    ):
        runner = BlackRunner()
        result = runner.invoke(
            main, args, input=BytesIO(input.encode("utf8")) if input else None
        )
        self.assertEqual(
            result.exit_code,
            exit_code,
            msg=(
                f"Failed with args: {args}\n"
                f"stdout: {runner.stdout_bytes.decode()!r}\n"
                f"stderr: {runner.stderr_bytes.decode()!r}\n"
                f"exception: {result.exception}"
            ),
        )
        return runner

    def test_no_flag(self):
        runner = self.invoke_curlylint(0, [])
        self.assertEqual(runner.stdout_bytes.decode(), "")
        self.assertEqual(
            runner.stderr_bytes.decode(), "No Path provided. Nothing to do 😴\n"
        )

    def test_stdin(self):
        runner = self.invoke_curlylint(0, ["-"], input="<p>Hello, World!</p>")
        self.assertEqual(runner.stdout_bytes.decode(), "")
        self.assertEqual(runner.stderr_bytes.decode(), "All done! ✨ 🍰 ✨\n\n")

    def test_stdin_verbose(self):
        runner = self.invoke_curlylint(
            0, ["--verbose", "-"], input="<p>Hello, World!</p>"
        )
        self.assertEqual(runner.stdout_bytes.decode(), "")
        self.assertIn(
            "Identified project root as:", runner.stderr_bytes.decode()
        )
        self.assertIn(
            """Analyzing file content from stdin
Files being analyzed:
-
All done! ✨ 🍰 ✨
""",
            runner.stderr_bytes.decode(),
        )

    def test_flag_help(self):
        runner = self.invoke_curlylint(0, ["--help"])
        self.assertIn(
            "Prototype linter for Jinja and Django templates",
            runner.stdout_bytes.decode(),
        )
        self.assertEqual(runner.stderr_bytes.decode(), "")

    def test_template_tags_validation_fail_no_nesting(self):
        runner = self.invoke_curlylint(
            2,
            ["--template-tags", '["cache", "endcache"]', "-"],
            input="<p>Hello, World!</p>",
        )
        self.assertIn(
            "Error: Invalid value for '--template-tags': expected a list of lists of tags as JSON, got '[\"cache\", \"endcache\"]'",
            runner.stderr_bytes.decode(),
        )

    def test_template_tags_cli_configured(self):
        self.invoke_curlylint(
            0,
            ["--template-tags", '[["of", "elseof", "endof"]]', "-"],
            input="<p>{% of a %}c{% elseof %}test{% endof %}</p>",
        )

    def test_template_tags_cli_unconfigured_fails(self):
        runner = self.invoke_curlylint(
            1,
            ["--template-tags", "[]", "-"],
            input="<p>{% of a %}c{% elseof %}test{% endof %}</p>",
        )
        self.assertIn(
            "Parse error: expected one of 'autoescape', 'block', 'blocktrans', 'comment', 'filter', 'for', 'if', 'ifchanged', 'ifequal', 'ifnotequal', 'not an intermediate Jinja tag name', 'spaceless', 'verbatim', 'with' at 0:17\tparse_error",
            runner.stdout_bytes.decode(),
        )



import unittest

from .issue import Issue, IssueLocation


class TestIssue(unittest.TestCase):
    def test_from_ast(self):
        self.assertEqual(
            Issue.from_ast(
                file="test.html",
                line=4,
                column=23,
                code="test_code",
                message="Test message",
            ),
            Issue(
                IssueLocation(file_path="test.html", line=4, column=23),
                "Test message",
                "test_code",
            ),
        )

    def test_from_dict(self):
        self.assertEqual(
            Issue.from_dict(
                {
                    "file": "test.html",
                    "line": 4,
                    "column": 23,
                    "code": "test_code",
                    "message": "Test message",
                }
            ),
            Issue.from_ast(
                file="test.html",
                line=4,
                column=23,
                code="test_code",
                message="Test message",
            ),
        )



from typing import List

from curlylint.issue import Issue


def format_compact(issues: List[Issue]):
    sorted_issues = sorted(
        issues,
        key=lambda i: (
            i.location.file_path,
            i.location.line,
            i.location.column,
        ),
    )

    for issue in sorted_issues:
        print(str(issue))



import re
from functools import partial
from pathlib import Path
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Pattern,
    Set,
    Tuple,
    Union,
)

import click  # lgtm [py/import-and-import-from]

from curlylint.rule_param import RULE
from curlylint.template_tags_param import TEMPLATE_TAGS

from . import __version__
from .config import (
    find_project_root,
    gen_template_files_in_dir,
    get_gitignore,
    read_pyproject_toml,
)
from .lint import lint, lint_one
from .report import Report

out = partial(click.secho, bold=True, err=True)
err = partial(click.secho, fg="red", err=True)

DEFAULT_EXCLUDES = r"/(\.eggs|\.git|\.hg|\.mypy_cache|\.nox|\.tox|\.venv|venv|myvenv|\.svn|_build|buck-out|build|dist|coverage_html_report|node_modules)/"
DEFAULT_INCLUDES = r"\.(html|jinja|twig)$"


def re_compile_maybe_verbose(regex: str) -> Pattern[str]:
    """Compile a regular expression string in `regex`.
    If it contains newlines, use verbose mode.
    """
    if "\n" in regex:
        regex = "(?x)" + regex
    compiled: Pattern[str] = re.compile(regex)
    return compiled


def path_empty(
    src: Tuple[str, ...], quiet: bool, verbose: bool, ctx: click.Context
) -> None:
    """
    Exit if there is no `src` provided for formatting
    """
    if not src:
        if verbose or not quiet:
            out("No Path provided. Nothing to do 😴")
            ctx.exit(0)


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.version_option(version=__version__)
@click.option("-v", "--verbose", is_flag=True, help="Verbose mode.")
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    help=(
        "Don’t emit non-error messages to stderr. Errors are still emitted; "
        "silence those with 2>/dev/null."
    ),
)
@click.option(
    "--parse-only",
    is_flag=True,
    help="Don’t lint, check for syntax errors and exit.",
)
@click.option(
    "--print-config",
    is_flag=True,
    help="Print the configuration for the given file.",
    is_eager=True,
)
@click.option(
    "-f",
    "--format",
    type=click.Choice(("compact", "json", "stylish")),
    default="stylish",
    help=("Use a specific output format"),
    show_default=True,
)
@click.option(
    "--stdin-filepath",
    type=click.Path(
        exists=False,
        file_okay=True,
        dir_okay=False,
        readable=False,
        allow_dash=False,
    ),
    help=("Path to the file to pretend that stdin comes from."),
)
@click.option(
    "--include",
    type=str,
    default=DEFAULT_INCLUDES,
    help=(
        "A regular expression that matches files and directories that should be "
        "included on recursive searches. An empty value means all files are "
        "included regardless of the name. Use forward slashes for directories on "
        "all platforms (Windows, too).  Exclusions are calculated first, inclusions "
        "later."
    ),
    show_default=True,
)
@click.option(
    "--exclude",
    type=str,
    default=DEFAULT_EXCLUDES,
    help=(
        "A regular expression that matches files and directories that should be "
        "excluded on recursive searches. An empty value means no paths are excluded. "
        "Use forward slashes for directories on all platforms (Windows, too).  "
        "Exclusions are calculated first, inclusions later."
    ),
    show_default=True,
)
@click.option(
    "--rule",
    type=RULE,
    help=(
        'Specify rules, with the syntax --rule \'code: {"json": "value"}\'. '
        "Can be provided multiple times to configure multiple rules."
    ),
    multiple=True,
)
@click.option(
    "--template-tags",
    type=TEMPLATE_TAGS,
    default="[]",
    help=(
        'Specify additional sets of template tags, with the syntax --template-tags \'[["cache", "endcache"]]\'. '
    ),
    show_default=True,
)
@click.argument(
    "src",
    nargs=-1,
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=True,
        readable=True,
        allow_dash=True,
    ),
    is_eager=True,
)
@click.option(
    "-c",
    "--config",
    type=click.Path(
        exists=False,
        file_okay=True,
        dir_okay=False,
        readable=True,
        allow_dash=False,
    ),
    is_eager=True,
    help="Read configuration from FILE.",
    callback=read_pyproject_toml,
)
@click.pass_context
def main(
    ctx: click.Context,
    verbose: bool,
    quiet: bool,
    parse_only: bool,
    print_config: bool,
    config: Optional[str],
    format: str,
    stdin_filepath: str,
    include: str,
    exclude: str,
    rule: Union[Mapping[str, Any], Tuple[Mapping[str, Any], ...]],
    template_tags: List[List[str]],
    src: Tuple[str, ...],
) -> None:
    """Prototype linter for Jinja and Django templates, forked from jinjalint"""

    if config and verbose:
        out(f"Using configuration from: {config}", bold=False, fg="blue")

    try:
        include_regex = re_compile_maybe_verbose(include)
    except re.error:
        err(f"Invalid regular expression for include given: {include!r}")
        ctx.exit(2)
    try:
        exclude_regex = re_compile_maybe_verbose(exclude)
    except re.error:
        err(f"Invalid regular expression for exclude given: {exclude!r}")
        ctx.exit(2)

    report = Report(quiet=quiet, verbose=verbose, format=format)
    root = find_project_root(src)

    if verbose:
        out(f"Identified project root as: {root}", bold=False, fg="blue")

    path_empty(src, quiet, verbose, ctx)

    sources: Set[Path] = set()
    for s in src:
        p = Path(s)
        if p.is_dir():
            sources.update(
                gen_template_files_in_dir(
                    p,
                    root,
                    include_regex,
                    exclude_regex,
                    report,
                    get_gitignore(root),
                )
            )
        elif p.is_file() or s == "-":
            if verbose:
                out("Analyzing file content from stdin")

            sources.add(p)
        else:
            err(f"invalid path: {s}")

    if len(sources) == 0:
        if verbose or not quiet:
            out(
                "No template files are present to be formatted. Nothing to do 😴"
            )
        ctx.exit(0)

    if verbose:
        out("Files being analyzed:")
        out("\n".join(str(s) for s in sources), bold=False, fg="blue")

    configuration = {}  # type: Dict[str, Any]
    if ctx.default_map:
        configuration.update(ctx.default_map)

    rules = configuration.get("rules", {})

    if rule:
        if isinstance(rule, tuple):
            for r in rule:
                rules.update(r)
        else:
            rules.update(rule)

    configuration["rules"] = rules
    configuration["verbose"] = verbose
    configuration["parse_only"] = parse_only
    configuration["template_tags"] = template_tags

    if stdin_filepath:
        configuration["stdin_filepath"] = Path(stdin_filepath)

    if len(sources) == 1:
        issues = lint_one(sources.pop(), configuration)
    else:
        issues = lint(sources, configuration)

    report.print_issues(issues)

    if verbose or not quiet:
        out("Oh no! 💥 💔 💥" if report.return_code == 1 else "All done! ✨ 🍰 ✨")
        click.secho(str(report), err=True)
    ctx.exit(report.return_code)


def patch_click() -> None:
    """Borrowed from black.
    https://github.com/psf/black/blob/959848c17639bfc646128f6b582c5858164a5001/black.py
    Make Click not crash.
    On certain misconfigured environments, Python 3 selects the ASCII encoding as the
    default which restricts paths that it can access during the lifetime of the
    application.  Click refuses to work in this scenario by raising a RuntimeError.
    In case of Black the likelihood that non-ASCII characters are going to be used in
    file paths is minimal since it's Python source code.  Moreover, this crash was
    spurious on Python 3.7 thanks to PEP 538 and PEP 540.
    """
    try:
        from click import core
        from click import _unicodefun  # type: ignore
    except ModuleNotFoundError:
        return

    for module in (core, _unicodefun):
        if hasattr(module, "_verify_python3_env"):
            module._verify_python3_env = lambda: None


def patched_main() -> None:
    patch_click()
    main()


if __name__ == "__main__":
    patched_main()



#!/usr/bin/env python

import io

from setuptools import find_packages, setup  # type: ignore

from curlylint import (
    __author__,
    __author_email__,
    __copyright__,
    __description__,
    __license__,
    __name__,
    __url__,
    __version__,
)

with io.open("README.md", encoding="utf-8") as readme_file:
    long_description = readme_file.read()

setup(
    name=__name__,
    version=__version__,
    description=__description__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=__url__,
    author=__author__,
    author_email=__author_email__,
    license=__license__,
    copyright=__copyright__,
    packages=find_packages(exclude=["tests*"]),
    include_package_data=True,
    package_data={"curlylint": ["py.typed"]},
    keywords=[
        "curlylint",
        "linter",
        "linting",
        "jinja",
        "jinja2",
        "django templates",
        "twig",
        "nunjucks",
        "liquid",
        "templates",
        "html",
    ],
    python_requires=">=3.6",
    install_requires=[
        "parsy==1.1.0",
        "attrs>=17.2.0",
        "click>=6.5",
        "toml>=0.9.4",
        "pathspec>=0.6, <1",
        "dataclasses>=0.6; python_version < '3.7'",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Quality Assurance",
        "Typing :: Typed",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    entry_points={"console_scripts": ["curlylint=curlylint.cli:patched_main"]},
)



from curlylint.check_node import CheckNode, build_tree
from curlylint.issue import Issue

ARIA_ROLE = "aria_role"

RULE = {
    "id": "aria_role",
    "type": "accessibility",
    "docs": {
        "description": "Elements with ARIA roles must use a valid, non-abstract ARIA role",
        "url": "https://www.curlylint.org/docs/rules/aria_role",
        "impact": "Serious",
        "tags": ["cat.language", "wcag2a", "wcag412"],
        "resources": [
            "[WAI-ARIA Role Definitions](https://www.w3.org/TR/wai-aria/#role_definitions)",
            "[WCAG2.1 SC 4.2.1: Name, Role, Value (Level A)](https://www.w3.org/WAI/WCAG21/Understanding/name-role-value)",
            "[eslint-plugin-jsx-a11y, aria-role](https://github.com/evcohen/eslint-plugin-jsx-a11y/blob/master/docs/rules/aria-role.md)",
        ],
    },
    "schema": {
        "$schema": "http://json-schema.org/draft/2019-09/schema#",
        "oneOf": [
            {
                "const": True,
                "title": "All role attributes must be valid.",
                "examples": [True],
            },
            {
                "type": "array",
                "items": {"type": "string"},
                "uniqueItems": True,
                "title": "All role attributes must match the provided list",
                "examples": [["search", "presentation", "alert"]],
            },
        ],
    },
}

VALID_ROLES = (
    # Abstract roles, which shouldn’t ever be used in web pages – included here for reference.
    # "command",
    # "composite",
    # "input",
    # "landmark",
    # "range",
    # "roletype",
    # "section",
    # "sectionhead",
    # "select",
    # "structure",
    # "widget",
    # "window",
    "alert",
    "alertdialog",
    "application",
    "article",
    "banner",
    "button",
    "cell",
    "checkbox",
    "columnheader",
    "combobox",
    "complementary",
    "contentinfo",
    "definition",
    "dialog",
    "directory",
    "document",
    "feed",
    "figure",
    "form",
    "grid",
    "gridcell",
    "group",
    "heading",
    "img",
    "link",
    "list",
    "listbox",
    "listitem",
    "log",
    "main",
    "marquee",
    "math",
    "menu",
    "menubar",
    "menuitem",
    "menuitemcheckbox",
    "menuitemradio",
    "navigation",
    "none",
    "note",
    "option",
    "presentation",
    "progressbar",
    "radio",
    "radiogroup",
    "region",
    "row",
    "rowgroup",
    "rowheader",
    "scrollbar",
    "search",
    "searchbox",
    "separator",
    "slider",
    "spinbutton",
    "status",
    "switch",
    "tab",
    "table",
    "tablist",
    "tabpanel",
    "term",
    "textbox",
    "timer",
    "toolbar",
    "tooltip",
    "tree",
    "treegrid",
    "treeitem",
    "doc-abstract",
    "doc-acknowledgments",
    "doc-afterword",
    "doc-appendix",
    "doc-backlink",
    "doc-biblioentry",
    "doc-bibliography",
    "doc-biblioref",
    "doc-chapter",
    "doc-colophon",
    "doc-conclusion",
    "doc-cover",
    "doc-credit",
    "doc-credits",
    "doc-dedication",
    "doc-endnote",
    "doc-endnotes",
    "doc-epigraph",
    "doc-epilogue",
    "doc-errata",
    "doc-example",
    "doc-footnote",
    "doc-foreword",
    "doc-glossary",
    "doc-glossref",
    "doc-index",
    "doc-introduction",
    "doc-noteref",
    "doc-notice",
    "doc-pagebreak",
    "doc-pagelist",
    "doc-part",
    "doc-preface",
    "doc-prologue",
    "doc-pullquote",
    "doc-qna",
    "doc-subtitle",
    "doc-tip",
    "doc-toc",
)


def find_role_errors(node, file, target_roles):
    attributes = []
    if getattr(node.value, "opening_tag", None):
        attributes = {}
        for n in node.value.opening_tag.attributes.nodes:
            attributes[str(n.name)] = str(n.value).strip("\"'")

    target = VALID_ROLES if target_roles is True else target_roles

    if "role" in attributes:
        return (
            []
            if attributes["role"] in target
            else [
                Issue.from_node(
                    file,
                    node,
                    "The `role` attribute needs to have a valid value",
                    "aria_role",
                )
            ]
        )

    if not node.children:
        return []

    return sum(
        (
            find_role_errors(child, file, target_roles)
            for child in node.children
        ),
        [],
    )


def aria_role(file, target_roles):
    root = CheckNode(None)
    build_tree(root, file.tree)
    src = file.source.lower()

    if r" role" in src:
        return find_role_errors(root, file, target_roles)

    return []



from curlylint import ast
from curlylint.check_node import CheckNode, build_tree
from curlylint.issue import Issue

META_VIEWPORT = "meta_viewport"

RULE = {
    "id": "meta_viewport",
    "type": "accessibility",
    "docs": {
        "description": "The `viewport` meta tag should not use `user-scalable=no`, and `maximum-scale` should be 2 or above, so end users can zoom",
        "url": "https://www.curlylint.org/docs/rules/meta_viewport",
        "impact": "Critical",
        "tags": ["cat.language", "wcag2aa", "wcag144"],
        "resources": [
            "[Understanding WCAG SC 1.4.4 Resize Text](http://www.w3.org/TR/UNDERSTANDING-WCAG20/visual-audio-contrast-scale.html)",
            "[axe-core, meta-viewport](https://dequeuniversity.com/rules/axe/3.5/meta-viewport)",
        ],
    },
    "schema": {
        "$schema": "http://json-schema.org/draft/2019-09/schema#",
        "oneOf": [
            {
                "const": True,
                "title": "`user-scalable=no` must not be used, and `maximum-scale` should be 2 or above.",
                "examples": [True],
            },
        ],
    },
}


def find_valid(node, file):
    name = getattr(node.value, "name", None)
    is_meta = (
        isinstance(node.value, ast.Element) and name and name.lower() == "meta"
    )

    if is_meta:
        attributes = []
        if getattr(node.value, "opening_tag", None):
            attributes = {}
            for n in node.value.opening_tag.attributes.nodes:
                attributes[str(n.name)] = str(n.value).strip("\"'")

        if "name" in attributes and attributes["name"] == "viewport":
            if "user-scalable=no" in attributes["content"]:
                return [
                    Issue.from_node(
                        file,
                        node,
                        "Remove `user-scalable=no` from the viewport meta so users can zoom",
                        "meta_viewport",
                    )
                ]

            if (
                "maximum-scale=1" in attributes["content"]
                or "maximum-scale=0" in attributes["content"]
            ):
                return [
                    Issue.from_node(
                        file,
                        node,
                        "`maximum-scale` should not be less than 2",
                        "meta_viewport",
                    )
                ]

    if not node.children:
        return []

    return sum((find_valid(child, file) for child in node.children), [])


def meta_viewport(file, config):
    root = CheckNode(None)
    build_tree(root, file.tree)
    src = file.source.lower()

    if r"user-scalable" in src or r"maximum-scale" in src:
        return find_valid(root, file)

    return []



import unittest

from curlylint.lint import parse_source
from curlylint.rules.rule_test_case import RulesTestMeta

from .html_has_lang import html_has_lang


class TestRule(unittest.TestCase, metaclass=RulesTestMeta):
    fixtures = __file__.replace(".py", ".json")
    rule = html_has_lang

    def test_skips(self):
        errors, file = parse_source("test.html", {}, "<p>Test</p>")
        self.assertEqual(html_has_lang(file, True), [])
